/* RunVCS - Configuration parser
 * Copyright (C) 2009, Daniel Collins <solemnwarning@solemnwarning.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *	* Redistributions of source code must retain the above copyright notice,
 *	  this list of conditions and the following disclaimer.
 *	
 *	* Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 *	
 *	* Neither the name of the developer nor the names of its contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ''AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE DEVELOPER BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <string>
#include <vector>

#include "runvcs.hpp"

typedef std::vector<std::string> token_list;

struct file {
	FILE *fh;
	
	std::string name;
	int line_num;
	
	std::string line_text;
	token_list line_toks;
};

static void lexer(struct file &file);
static int next_line(struct file &file);
static void do_line(struct file &file, int enable);
static void do_if(struct file &file, int enable);
static void allow_svn(struct file &file, int enable);

void load_config(std::string file) {
	struct file fs;
	
	fs.fh = fopen(file.c_str(), "r");
	if(!fs.fh) {
		die("Error opening %s: %s", file.c_str(), strerror(errno));
	}
	
	fs.name = file;
	fs.line_num = 0;
	
	while(next_line(fs)) {
		do_line(fs, 1);
	}
	
	fclose(fs.fh);
}

static void lexer(struct file &file) {
	char qchar = '\0';
	std::string token = "";
	char const *data = file.line_text.c_str();
	
	file.line_toks.clear();
	
	for(; *data; data++) {
		if(*data == '\\') {
			token.append(1, *(++data));
			continue;
		}
		
		if(qchar) {
			if(*data == qchar) {
				file.line_toks.push_back(token);
				token.clear();
				
				qchar = '\0';
			}else{
				token.append(1, *data);
			}
			
			continue;
		}
		if(strchr("'\"", *data)) {
			qchar = *data;
			continue;
		}
		
		if(strchr(" \t\r\n{}=,", *data)) {
			if(!token.empty()) {
				file.line_toks.push_back(token);
				token.clear();
			}
			
			if(strchr("{}=,", *data)) {
				file.line_toks.push_back(std::string(1, *data));
			}
		}else{
			token.append(1, *data);
		}
	}
	
	if(qchar) {
		die("%s:%d: Unmatched %c", file.name.c_str(), file.line_num, qchar);
	}
	
	if(!token.empty()) {
		file.line_toks.push_back(token);
	}
}

static int next_line(struct file &file) {
	char buf[1024];
	
	if(fgets(buf, sizeof(buf), file.fh)) {
		file.line_num++;
		file.line_text = buf;
		lexer(file);
		
		return 1;
	}
	
	if(ferror(file.fh)) {
		die("%s: %s", file.name.c_str(), strerror(errno));
	}
	
	return 0;
}

static void do_line(struct file &file, int enable) {
	if(file.line_toks.empty()) {
		return;
	}
	
	std::string s = file.line_toks.at(0);
	
	if(s == "if" || s == "!if") {
		do_if(file, enable);
	}else if(s == "endif") {
		die("%s:%d: Unmatched 'endif'", file.name.c_str(), file.line_num);
	}else if(s == "allow-svn") {
		allow_svn(file, enable);
	}else{
		die("%s:%d: Unknown directive '%s'", file.name.c_str(), file.line_num, s.c_str());
	}
}

static void do_if(struct file &file, int enable) {
	int state = 0, line = file.line_num;
	std::string name = file.name;
	
	if(file.line_toks.size() != 3) {
		die("%s:%d: 'if' expects 2 arguments", name.c_str(), line);
	}
	
	std::string test = file.line_toks.at(1);
	std::string val = file.line_toks.at(2);
	
	if(test == "profile") {
		std::vector<std::string>::iterator profile = profiles.begin();
		std::vector<std::string>::iterator pend = profiles.end();
		
		while(profile != pend) {
			if(*profile == val) {
				state = 1;
				break;
			}
			
			profile++;
		}
	}else if(test == "user") {
		if(strcasecmp(username.c_str(), val.c_str()) == 0) {
			state = 1;
		}
	}else if(test == "uid") {
		if(strtoul(val.c_str(), NULL, 10) == getuid()) {
			state = 1;
		}
	}else{
		die("%s:%d: Unknown test '%s'", name.c_str(), line, test.c_str());
	}
	
	if(file.line_toks.at(0) == "!if") {
		state = (state ? 0 : 1);
	}
	if(!enable) {
		state = 0;
	}
	
	while(next_line(file)) {
		std::string t0 = file.line_toks.at(0);
		
		if(t0 == "endif") {
			return;
		}
		
		do_line(file, state);
	}
	
	die("%s:%d: Missing endif", name.c_str(), line);
}

static void allow_svn(struct file &file, int enable) {
	int i = file.line_toks.size();
	
	if(!(i == 2 || (i == 4 && file.line_toks.at(2) == "="))) {
		die("%s:%d: Invalid arguments", file.name.c_str(), file.line_num);
	}
	
	if(enable) {
		struct svn_repo repo;
		repo.name = file.line_toks.at(1);
		repo.path = (i == 4 ? file.line_toks.at(3) : repo.name);
		
		svn_repos.push_back(repo);
	}
}
