/* RunVCS - Main
 * Copyright (C) 2009, Daniel Collins <solemnwarning@solemnwarning.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *	* Redistributions of source code must retain the above copyright notice,
 *	  this list of conditions and the following disclaimer.
 *	
 *	* Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 *	
 *	* Neither the name of the developer nor the names of its contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ''AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE DEVELOPER BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <dirent.h>
#include <pwd.h>
#include <string>
#include <vector>

#include "runvcs.hpp"

static std::string init_user(void);
static std::string init_homedir(void);

std::string username = init_user();
std::vector<std::string> profiles;
std::vector<struct svn_repo> svn_repos;
std::string homedir = init_homedir();

static void usage(char const *argv0);
static std::vector<std::string> parse_cmd(char const *line);
static int exec_cmd(std::vector<std::string> args);
static int exec_svn(void);
static char *svn_mkroot(void);
static int mkdirs(char const *root, char const *dirs);
static void cleanup_tmp(void);
static int rmtmp(char const *path);

int main(int argc, char **argv) {
	int arg;
	char *line = NULL, *oc = NULL;
	opterr = 0;
	
	while((arg = getopt(argc, argv, "u:p:c:v")) != -1) {
		if(arg == 'u') {
			username = optarg;
			
			setenv("LOGNAME", username.c_str(), 1);
			setenv("USER", username.c_str(), 1);
		}
		if(arg == 'p') {
			profiles.push_back(std::string(optarg));
		}
		if(arg == '?') {
			usage(argv[0]);
		}
		if(arg == 'c') {
			line = optarg;
		}
		if(arg == 'v') {
			printf("RunVCS " VERSION " - Copyright (C) 2009 Daniel Collins\n");
			printf("Configuration files: " ETCDIR "/runvcs.conf, %s/.runvcs\n", homedir.c_str());
			return 0;
		}
	}
	if(optind != argc) {
		usage(argv[0]);
	}
	
	if(access(ETCDIR "/runvcs.conf", F_OK) == 0) {
		load_config(ETCDIR "/runvcs.conf");
	}
	if(homedir != "???") {
		std::string hconfig = homedir + "/.runvcs";
		
		if(access(hconfig.c_str(), F_OK) == 0) {
			load_config(hconfig);
		}
	}
	
	if((oc = getenv("SSH_ORIGINAL_COMMAND"))) {
		line = oc;
	}
	
	if(line) {
		std::vector<std::string> args = parse_cmd(line);
		int ret = exec_cmd(args);
		
		cleanup_tmp();
		return ret;
	}
	
	while((line = readline("runvcs:$ "))) {
		if(line[strspn(line, " ")] == '\0') {
			free(line);
			continue;
		}
		
		std::vector<std::string> args = parse_cmd(line);
		
		add_history(line);
		free(line);
		
		if(args.at(0) == "exit") {
			break;
		}
		
		exec_cmd(args);
	}
	
	cleanup_tmp();
	
	return 0;
}

static void usage(char const *argv0) {
	printf("Usage: %s [-u <username>] [-p <profile>] [-s]\n", argv0);
	printf("-u <username>\tSet username\n");
	printf("-p <profile>\tEnable a profile name, can be used multiple times\n");
	printf("-v\t\t\tPrint version information and exit\n");
	exit(1);
}

void die(char const *fmt, ...) {
	char msgbuf[1024];
	va_list argv;
	
	va_start(argv, fmt);
	vsnprintf(msgbuf, 1024, fmt, argv);
	va_end(argv);
	
	fprintf(stderr, "%s\n", msgbuf);
	
	cleanup_tmp();
	
	exit(1);
}

static std::vector<std::string> parse_cmd(char const *line) {
	int qchar = '\0';
	std::vector<std::string> args;
	std::string arg = "";
	
	for(; *line; line++) {
		if(*line == '\\') {
			arg.append(1, *(++line));
			continue;
		}
		
		if(*line == qchar) {
			qchar = '\0';
			continue;
		}
		if(strchr("'\"", *line)) {
			qchar = *line;
			continue;
		}
		
		if(*line == ' ' && !qchar) {
			if(!arg.empty()) {
				args.push_back(arg);
				arg.clear();
			}
		}else{
			arg.append(1, *line);
		}
	}
	
	if(!arg.empty()) {
		args.push_back(arg);
	}
	
	return args;
}

static int exec_cmd(std::vector<std::string> args) {
	if(args.empty()) {
		return 0;
	}
	
	if(args.at(0) == "svnserve") {
		if(args.size() != 2 || (args.at(1) != "-t" && args.at(1) != "--tunnel")) {
			fprintf(stderr, "svnserve must be called in tunnel mode\n");
			return 1;
		}
		
		return exec_svn();
	}
	
	fprintf(stderr, "Unknown command '%s'\n", args.at(0).c_str());
	return 1;
}

static int exec_svn(void) {
	char *svn_root = svn_mkroot();
	if(!svn_root) {
		return 1;
	}
	
	int svnpid = fork();
	if(svnpid == -1) {
		fprintf(stderr, "fork: %s\n", strerror(errno));
		delete svn_root;
		
		return 1;
	}
	if(svnpid == 0) {
		std::string user_arg = "--tunnel-user=";
		user_arg.append(username);
		
		execlp("svnserve", "svnserve", "-t", "-r", svn_root, user_arg.c_str(), NULL);
		
		fprintf(stderr, "svnserve: %s\n", strerror(errno));
		exit(1);
	}
	
	delete svn_root;
	
	int status;
	wait(&status);
	
	return WEXITSTATUS(status);
}

static char *svn_mkroot(void) {
	char *root = new char[128];
	sprintf(root, "/tmp/runvcs.%d/svnroot/a", (int)getpid());
	
	if(!mkdirs("/", root)) {
		delete root;
		return NULL;
	}
	
	strrchr(root, '/')[1] = '\0';
	
	std::vector<struct svn_repo>::iterator repo = svn_repos.begin();
	std::vector<struct svn_repo>::iterator end = svn_repos.end();
	
	while(repo != end) {
		char const *rname = repo->name.c_str();
		std::string lname = root;
		lname.append(rname);
		
		if(!mkdirs(root, rname)) {
			delete root;
			return NULL;
		}
		
		if(symlink(repo->path.c_str(), lname.c_str()) == -1 && errno != EEXIST) {
			fprintf(stderr, "Failed to symlink %s: %s\n", lname.c_str(), strerror(errno));
			
			delete root;
			return NULL;
		}
		
		repo++;
	}
	
	return root;
}

static int mkdirs(char const *root, char const *dirs) {
	std::string dir = root;
	
	while(*dirs) {
		int len = strcspn(dirs, "/");
		
		dir.append(dirs, len);
		dir.append(1, '/');
		
		dirs += len;
		dirs += strspn(dirs, "/");
		
		if(*dirs && len && mkdir(dir.c_str(), 0755) == -1 && errno != EEXIST) {
			fprintf(stderr, "Failed to mkdir %s: %s\n", dir.c_str(), strerror(errno));
			return 0;
		}
	}
	
	return 1;
}

static void cleanup_tmp(void) {
	char tmpdir[128];
	snprintf(tmpdir, 128, "/tmp/runvcs.%d", (int)getpid());
	
	if(access(tmpdir, F_OK) == -1) {
		return;
	}
	
	rmtmp(tmpdir);
}

static int rmtmp(char const *path) {
	struct stat finfo;
	
	if(lstat(path, &finfo) == -1) {
		fprintf(stderr, "Failed to stat %s: %s\n", path, strerror(errno));
		return 0;
	}
	
	if(S_ISDIR(finfo.st_mode)) {
		DIR *dir = opendir(path);
		if(!dir) {
			fprintf(stderr, "Failed to open %s: %s\n", path, strerror(errno));
			return 0;
		}
		
		struct dirent *file;
		
		while((file = readdir(dir))) {
			if(!strcmp(file->d_name, ".") || !strcmp(file->d_name, "..")) {
				continue;
			}
			
			std::string fpath = path;
			fpath.append(1, '/');
			fpath.append(file->d_name);
			
			if(!rmtmp(fpath.c_str())) {
				closedir(dir);
				return 0;
			}
		}
		
		closedir(dir);
		
		if(rmdir(path) == -1) {
			fprintf(stderr, "Failed to rmdir %s: %s\n", path, strerror(errno));
			return 0;
		}
	}else{
		if(unlink(path) == -1) {
			fprintf(stderr, "Failed to unlink %s: %s\n", path, strerror(errno));
			return 0;
		}
	}
	
	return 1;
}

static std::string init_user(void) {
	char const *s;
	
	if((s = getenv("LOGNAME"))) {
		return s;
	}
	if((s = getenv("USER"))) {
		return s;
	}
	
	struct passwd *passwd = getpwuid(getuid());
	if(passwd) {
		return passwd->pw_name;
	}
	
	fprintf(stderr, "Unable to determine username: %s", strerror(errno));
	return "???";
}

static std::string init_homedir(void) {
	char const *s;
	
	if((s = getenv("HOME"))) {
		return s;
	}
	
	struct passwd *passwd = getpwuid(getuid());
	if(passwd) {
		return passwd->pw_dir;
	}
	
	fprintf(stderr, "Unable to determine homedir: %s", strerror(errno));
	return "???";
}
